@@ -1,5 +1,6 @@
-from typing import Dict, List, Optional, Type
+from typing import Dict, List, Optional, Tuple, Type, Union
 
+import astropy.units as u
 import lightweaver as lw
 import lightweaver.wittmann as witt
 import numpy as np
@@ -102,6 +103,16 @@ class PctrPromModel(PromModel):
         When running a grid of models, consider creating a
         `TabulatedPromBcProvider` using `compute_falc_bc_ctx` and `tabulate_bc`,
         since the default performs quite a few extra RT calculations.
+    add_vertical_ray : bool, optional
+        Whether to add a non-weighted vertical ray to the model. This doesn't
+        participate in energy balance but allows for directly extracting mu=1,
+        without needing to do a boundary condition. Default: False. Equivalent to
+        `add_extra_mus={"muz": [1.0], "mux": [0.0]}`.
+    add_extra_rays : dict, optional
+        Extra rays to add to the quadrature along which the solution should be
+        sampled. These rays will not have a weight for integration, but can
+        serve as output for different viewing angles. Should be a dict with keys
+        `muz` and `mux` as iterables.
     """
 
     def __init__(
@@ -130,6 +141,8 @@ class PctrPromModel(PromModel):
         BcType: Optional[Type[PromBc]] = None,
         bc_kwargs=None,
         bc_provider=None,
+        add_vertical_ray: bool = False,
+        add_extra_rays: Dict[str, Union[List[float], Tuple[float], np.ndarray]] = None,
     ):
         self.projection = projection
         if projection not in ["prominence", "filament"]:
@@ -160,6 +173,13 @@ class PctrPromModel(PromModel):
                 "Cannot set both vrad and vlos for a filament. (Just set one of the two)."
             )
 
+        if add_vertical_ray and add_extra_rays is not None:
+            raise ValueError(
+                "Cannot provide extra_rays and set extra_rays dict simultaneously."
+            )
+        if add_vertical_ray:
+            add_extra_rays = {"muz": [1.0], "mux": [0.0]}
+
         if projection == "filament" and vrad is not None and vlos is None:
             vlos = vrad
             self.vlos = vrad
@@ -227,9 +247,9 @@ class PctrPromModel(PromModel):
                 witt.BK * self.temperature[k]
             )
 
-        rho /= lw.CM_TO_M**3 / lw.G_TO_KG
+        rho = (rho << u.Unit("g / cm3")).to("kg / m3").value
         nHTot = rho / (lw.Amu * lw.DefaultAtomicAbundance.massPerH)
-        ne /= lw.CM_TO_M**3
+        ne = (ne << u.Unit("cm-3")).to("m-3").value
         # NOTE(cmo): CGS Ends Here
 
         starting_z = np.zeros_like(cmass)
@@ -237,6 +257,7 @@ class PctrPromModel(PromModel):
             starting_z[k] = starting_z[k - 1] + 2.0 * (cmass[k] - cmass[k - 1]) / (
                 rho[k] + rho[k - 1]
             )
+        starting_z = np.ascontiguousarray(starting_z[::-1])
 
         lower_bc = BcType(projection, bc_provider, altitude, "lower", **bc_kwargs)
         if projection == "prominence":
@@ -260,6 +281,14 @@ class PctrPromModel(PromModel):
         )
         atmos = self.atmos
         atmos.quadrature(Nrays)
+        if add_extra_rays is not None:
+            extra_wmu = np.zeros_like(add_extra_rays["muz"])
+            atmos.rays(
+                muz=np.concatenate([atmos.muz, add_extra_rays["muz"]]),
+                mux=np.concatenate([atmos.mux, add_extra_rays["mux"]]),
+                wmu=np.concatenate([atmos.wmu, extra_wmu])
+            )
+
         self.rad_set = lw.RadiativeSet(atomic_models)
         self.rad_set.set_active(*active_atoms)
         if detailed_atoms is not None:
@@ -332,7 +361,7 @@ class PctrPromModel(PromModel):
                 )
                 if printNow:
                     print(
-                        f"    nHTotError: {np.max(np.abs(nHTotCorrection / self.atmos.nHTot))}"
+                        f"    nHTotError: {np.max(np.abs(nHTotCorrection / self.atmos.nHTot)):6.4e}"
                     )
                 self.atmos.ne[:] += (
                     nHTotCorrection * self.eq_pops["H"][-1] / self.atmos.nHTot
@@ -350,7 +379,7 @@ class PctrPromModel(PromModel):
                 # NOTE(cmo): The only iteration difference from the Iso case
                 rho = self.atmos.nHTot * lw.DefaultAtomicAbundance.massPerH * lw.Amu
                 for k in range(1, rho.shape[0]):
-                    self.atmos.z[k] = self.atmos.z[k - 1] + 2 * (
+                    self.atmos.z[k] = self.atmos.z[k - 1] - 2 * (
                         (self.cmass[k] - self.cmass[k - 1]) / (rho[k] + rho[k - 1])
                     )
                 # to here.
